!> @file ppm_graph_partition_mpi.f90
!! @brief generic graph partition method interface
!!
!! @copyright Copyright  (C)  2010  Thomas Jahns <jahns@dkrz.de>
!!
!! @version 1.0
!! @author Thomas Jahns <jahns@dkrz.de>
! Keywords: graph partitioning
! Maintainer: Thomas Jahns <jahns@dkrz.de>
! URL: https://www.dkrz.de/redmine/projects/show/scales-ppm
!
! Redistribution and use in source and binary forms, with or without
! modification, are  permitted provided that the following conditions are
! met:
!
! Redistributions of source code must retain the above copyright notice,
! this list of conditions and the following disclaimer.
!
! Redistributions in binary form must reproduce the above copyright
! notice, this list of conditions and the following disclaimer in the
! documentation and/or other materials provided with the distribution.
!
! Neither the name of the DKRZ GmbH nor the names of its contributors
! may be used to endorse or promote products derived from this software
! without specific prior written permission.
!
! THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
! IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
! TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
! PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
! OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
! EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
! PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
! PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
! LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
! NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
! SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
!
!> This is currently only a convenient wrapper of ParMeTis, other
!! heuristics are to follow later.
MODULE ppm_graph_partition_mpi
  USE iso_c_binding, ONLY: c_int, c_float
  USE ppm_base, ONLY: abort_ppm
  IMPLICIT NONE
  PRIVATE
  INCLUDE 'mpif.h'
#include <ppm.inc>
  INTERFACE
    SUBROUTINE parmetis_v3_partkway(vtxdist, xadj, adjncy, vwgt, adjwgt, &
         wgtflag, numflag, ncon, nparts, tpwgts, ubvec, options, edgecut, &
         part, comm)
      USE iso_c_binding, ONLY: c_int, c_float
      IMPORT :: ppm_idx
      INTEGER(ppm_idx), INTENT(in) :: vtxdist(*), xadj(*), adjncy(*), &
           vwgt(*), adjwgt(*)
      INTEGER(c_int), INTENT(in) :: wgtflag, numflag, ncon, nparts, options(*)
      REAL(c_float), INTENT(in) :: tpwgts(ncon, nparts), ubvec(ncon)
      INTEGER(c_int), INTENT(out) :: edgecut
      INTEGER(ppm_idx), INTENT(out) :: part(*)
      INTEGER, INTENT(in) :: comm
    END SUBROUTINE parmetis_v3_partkway
  END INTERFACE
  PUBLIC :: graph_partition_parmetis
CONTAINS
  !> call parmetis to partition graph in distributed CSR format
  !!
  !! @param num_vertices number of vertices
  !! @param edge_list_lens cumulative number of edges per vertex
  !! @param edge_lists concatenated list of vertices connected to each vertex
  !! @param partition_out for each vertex passed write partition assignment here
  !! @param comm optional communicator object identifying participating
  !! processes, defaults to ppm_default_comm
  !! @param num_partitions optional number of desired partitions,
  !! defaults to size of comm
  !! @param balance specify desired distribution of loads
  !! @param num_vertex_weights number of weighting factors per vertex
  !! @param vertex_weights
  SUBROUTINE graph_partition_parmetis(num_vertices, edge_list_lens, &
       edge_lists, partition_out, comm, num_partitions, &
       balance, num_vertex_weights, vertex_weights, edge_weights)
    INTEGER(ppm_idx), INTENT(in) :: num_vertices
    INTEGER(ppm_idx), INTENT(in) :: edge_list_lens(:)
    INTEGER(ppm_idx), INTENT(in) :: edge_lists(:)
    INTEGER(ppm_idx), INTENT(out) :: partition_out(*)
    INTEGER, OPTIONAL, INTENT(in) :: comm
    INTEGER, OPTIONAL, INTENT(in) :: num_partitions
    REAL(c_float), OPTIONAL, INTENT(in) :: balance(:, :)
    INTEGER, OPTIONAL, INTENT(in) :: num_vertex_weights
    INTEGER(ppm_idx), OPTIONAL, INTENT(in) :: vertex_weights(:)
    INTEGER(ppm_idx), OPTIONAL, INTENT(in) :: edge_weights(:)
    INTEGER(c_int) :: wgtflag
    INTEGER :: part_comm, comm_size, comm_rank, ierror, i
    INTEGER, ALLOCATABLE :: vtxdist(:)
    INTEGER(c_int) :: metis_options(0:2), edge_cut, num_parts
    INTEGER, PARAMETER :: max_msg_len=1024
    CHARACTER(len=max_msg_len) :: msg
    INTEGER(ppm_idx) :: dummy_weights(1)
    REAL(c_float) :: dummy_balance(1)

    IF (PRESENT(comm)) THEN; part_comm = comm; ELSE; part_comm = mpi_comm_world
    END IF
    CALL mpi_comm_size(part_comm, comm_size, ierror)
    IF (ierror /= MPI_SUCCESS) THEN
      CALL mpi_error_string(ierror, msg, max_msg_len, ierror)
      CALL abort_ppm(msg, __FILE__, __LINE__, comm)
    END IF
    CALL mpi_comm_rank(part_comm, comm_rank, ierror)
    IF (ierror /= MPI_SUCCESS) THEN
      CALL mpi_error_string(ierror, msg, max_msg_len, ierror)
      CALL abort_ppm(msg, __FILE__, __LINE__, comm)
    END IF

    ! build table of node distribution
    ALLOCATE(vtxdist(0:comm_size))

    i = INT(num_vertices)
    CALL mpi_allgather(i, 1, MPI_INTEGER, &
         vtxdist(1:comm_size), 1, MPI_INTEGER, part_comm, ierror)
    IF (ierror /= MPI_SUCCESS) then
      CALL mpi_error_string(ierror, msg, max_msg_len, ierror)
      CALL abort_ppm(msg, __FILE__, __LINE__, comm)
    END IF

    vtxdist(0) = 1
    vtxdist(comm_rank + 1) = i
    DO i = 1, comm_size
      vtxdist(i) = vtxdist(i) + vtxdist(i - 1)
    END DO

    wgtflag = 0
    IF (PRESENT(vertex_weights)) wgtflag = 2
    IF (PRESENT(edge_weights)) wgtflag = IOR(wgtflag, 1)
    IF (PRESENT(num_partitions)) THEN
      num_parts = num_partitions
    ELSE
      num_parts = comm_size
    END IF
    metis_options(0) = 0
    IF (PRESENT(balance) .AND. PRESENT(edge_weights)) THEN
      CALL parmetis_v3_partkway(vtxdist, edge_list_lens, edge_lists, &
           vertex_weights, edge_weights, wgtflag, 1, num_vertex_weights, &
           num_parts, balance, (/ REAL(1.05, c_float) /), metis_options, &
           edge_cut, partition_out, part_comm)
    ELSE IF(PRESENT(balance)) THEN
      CALL parmetis_v3_partkway(vtxdist, edge_list_lens, edge_lists, &
           vertex_weights, dummy_weights, wgtflag, 1, num_vertex_weights, &
           num_parts, balance, (/ REAL(1.05, c_float) /), metis_options, &
           edge_cut, partition_out, part_comm)
    ELSE ! neighter balance nor edge_weights present
      CALL parmetis_v3_partkway(vtxdist, edge_list_lens, edge_lists, &
           vertex_weights, dummy_weights, wgtflag, 1, num_vertex_weights, &
           num_parts, dummy_balance, (/ REAL(1.05, c_float) /), metis_options, &
           edge_cut, partition_out, part_comm)
    END IF
  END SUBROUTINE graph_partition_parmetis
END MODULE ppm_graph_partition_mpi
!
! Local Variables:
! license-project-url: "https://www.dkrz.de/redmine/projects/show/scales-ppm"
! license-default: "bsd"
! End:
