! ppm_graph_partition_serial.f90 --- wrapper for single task graph partitioner
!
! Copyright  (C)  2011  Thomas Jahns <jahns@dkrz.de>
!
! Version: 1.0
! Keywords: graph partitioning
! Author: Thomas Jahns <jahns@dkrz.de>
! Maintainer: Thomas Jahns <jahns@dkrz.de>
! URL: https://www.dkrz.de/redmine/projects/show/scales-ppm
!
! Redistribution and use in source and binary forms, with or without
! modification, are  permitted provided that the following conditions are
! met:
!
! Redistributions of source code must retain the above copyright notice,
! this list of conditions and the following disclaimer.
!
! Redistributions in binary form must reproduce the above copyright
! notice, this list of conditions and the following disclaimer in the
! documentation and/or other materials provided with the distribution.
!
! Neither the name of the DKRZ GmbH nor the names of its contributors
! may be used to endorse or promote products derived from this software
! without specific prior written permission.
!
! THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
! IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
! TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
! PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
! OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
! EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
! PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
! PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
! LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
! NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
! SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
!
!> perform partitioning of graph from serial code
MODULE ppm_graph_partition_serial
  USE iso_c_binding, ONLY: c_int, c_float
  USE ppm_base, ONLY: assertion
  USE ppm_extents, ONLY: extent
  USE ppm_graph_csr, ONLY: graph_csr, num_nodes
  USE ppm_set_partition_base, ONLY: partition_assignment
  USE ppm_std_type_kinds, ONLY: i4
  IMPLICIT NONE
  PRIVATE
#include <ppm.inc>
  EXTERNAL :: METIS_mCPartGraphKway
  EXTERNAL :: METIS_PartGraphKway
  PUBLIC :: graph_partition_metis
  INTERFACE graph_partition_metis
    MODULE PROCEDURE graph_partition_metis_base
    MODULE PROCEDURE graph_partition_metis_csr
    MODULE PROCEDURE graph_partition_metis_csr_mweights
  END INTERFACE graph_partition_metis
CONTAINS
  SUBROUTINE graph_partition_metis_base(num_vertices, edge_list_lens, &
       edge_lists, partition_out, num_partitions, &
       imbalance_tolerance, vertex_weights, edge_weights)
    INTEGER(ppm_idx), INTENT(in) :: num_vertices
    INTEGER(ppm_idx), INTENT(in) :: edge_list_lens(:)
    INTEGER(ppm_idx), INTENT(in) :: edge_lists(:)
    INTEGER(ppm_idx), INTENT(out) :: partition_out(:)
    INTEGER, INTENT(in) :: num_partitions
    REAL(c_float), OPTIONAL, INTENT(in) :: imbalance_tolerance(:)
    INTEGER(ppm_idx), OPTIONAL, INTENT(in) :: vertex_weights(*)
    INTEGER(ppm_idx), OPTIONAL, INTENT(in) :: edge_weights(*)
    INTEGER(ppm_idx) :: vw_dummy(1), ew_dummy(1)
    INTEGER(c_int) :: wgtflag
    INTEGER(c_int) :: metis_options(0:4), edge_cut

    metis_options(0) = 0
    metis_options(1:4) = 0
    IF (PRESENT(imbalance_tolerance)) THEN
      CALL assertion(PRESENT(vertex_weights), line=__LINE__, &
           source=__FILE__, &
           msg="when imbalance_tolerance is provided, vertex weights&
           & are also required")
      wgtflag = MERGE(1, 0, PRESENT(edge_weights))
      IF (PRESENT(edge_weights)) THEN
        CALL metis_mCPartGraphKway(INT(num_vertices, c_int), &
             INT(SIZE(imbalance_tolerance), c_int), &
             edge_list_lens, edge_lists, &
             vertex_weights, edge_weights, wgtflag, 1_c_int, &
             INT(num_partitions, c_int), &
             imbalance_tolerance, metis_options, edge_cut, partition_out)
      ELSE
        CALL metis_mCPartGraphKway(INT(num_vertices, c_int), &
             INT(SIZE(imbalance_tolerance), c_int), &
             edge_list_lens, edge_lists, &
             vertex_weights, ew_dummy, wgtflag, 1_c_int, &
             INT(num_partitions, c_int), &
             imbalance_tolerance, metis_options, edge_cut, partition_out)
      END IF
    ELSE
      wgtflag = MERGE(2, 0, PRESENT(vertex_weights))
      wgtflag = IOR(wgtflag, MERGE(1, 0, PRESENT(edge_weights)))
      SELECT CASE(wgtflag)
      CASE(0)
        CALL MeTiS_PartGraphKway(INT(num_vertices, c_int), &
             edge_list_lens, edge_lists, &
             vw_dummy, ew_dummy, wgtflag, INT(1, c_int), &
             INT(num_partitions, c_int), &
             metis_options, edge_cut, partition_out)
      CASE(1)
        CALL MeTiS_PartGraphKway(INT(num_vertices, c_int), &
             edge_list_lens, edge_lists, &
             vw_dummy, edge_weights, wgtflag, INT(1, c_int), &
             INT(num_partitions, c_int), &
             metis_options, edge_cut, partition_out)
      CASE(2)
        CALL MeTiS_PartGraphKway(INT(num_vertices, c_int), &
             edge_list_lens, edge_lists, &
             vertex_weights, ew_dummy, wgtflag, INT(1, c_int), &
             INT(num_partitions, c_int), &
             metis_options, edge_cut, partition_out)
      CASE(3)
        CALL MeTiS_PartGraphKway(INT(num_vertices, c_int), &
             edge_list_lens, edge_lists, &
             vertex_weights, edge_weights, wgtflag, INT(1, c_int), &
             INT(num_partitions, c_int), &
             metis_options, edge_cut, partition_out)
      END SELECT
    END IF
  END SUBROUTINE graph_partition_metis_base

  SUBROUTINE graph_partition_metis_csr(partition, graph, num_partitions, &
       vertex_weights, edge_weights)
    TYPE(partition_assignment), INTENT(out) :: partition
    TYPE(graph_csr), INTENT(in) :: graph
    INTEGER, INTENT(in) :: num_partitions
    INTEGER(ppm_idx), OPTIONAL, INTENT(in) :: vertex_weights(:)
    INTEGER(ppm_idx), OPTIONAL, INTENT(in) :: edge_weights(:)

    INTEGER :: n, wgtflag
    INTEGER(ppm_idx), ALLOCATABLE :: xadj(:), adjncy(:)

    n = num_nodes(graph)
    wgtflag = MERGE(2, 0, PRESENT(vertex_weights))
    wgtflag = IOR(wgtflag, MERGE(1, 0, PRESENT(edge_weights)))
    partition%part_range = extent(1, num_partitions)
    ALLOCATE(partition%assigned(1:n))
    IF (ppm_idx == i4) THEN
      SELECT CASE(wgtflag)
      CASE(0)
        CALL graph_partition_metis_base(n, graph%edges_of_vtx, &
             graph%edges, partition%assigned, num_partitions)
      CASE(1)
        CALL graph_partition_metis_base(n, graph%edges_of_vtx, &
             graph%edges, partition%assigned, &
             num_partitions, edge_weights=edge_weights)
      CASE(2)
        CALL graph_partition_metis_base(n, graph%edges_of_vtx, &
             graph%edges, partition%assigned, &
             num_partitions, vertex_weights=vertex_weights)
      CASE(3)
        CALL graph_partition_metis_base(n, graph%edges_of_vtx, &
             graph%edges, partition%assigned, num_partitions, &
             vertex_weights=vertex_weights, edge_weights=edge_weights)
      END SELECT
    ELSE
      ALLOCATE(xadj(SIZE(graph%edges_of_vtx)), adjncy(SIZE(graph%edges)))
      xadj = INT(graph%edges_of_vtx, ppm_idx)
      adjncy = INT(graph%edges, ppm_idx)

      CALL graph_partition_metis_base(n, xadj, adjncy, partition%assigned, &
           num_partitions, vertex_weights=vertex_weights, &
           edge_weights=edge_weights)
    END IF
  END SUBROUTINE graph_partition_metis_csr

  SUBROUTINE graph_partition_metis_csr_mweights(partition, graph, &
       num_partitions, imbalance_tolerance, vertex_weights, edge_weights)
    TYPE(partition_assignment), INTENT(out) :: partition
    TYPE(graph_csr), INTENT(in) :: graph
    INTEGER, INTENT(in) :: num_partitions
    REAL(c_float), INTENT(in) :: imbalance_tolerance(:)
    INTEGER(ppm_idx), INTENT(in) :: vertex_weights(:,:)
    INTEGER(ppm_idx), OPTIONAL, INTENT(in) :: edge_weights(:)

    INTEGER :: n
    INTEGER(ppm_idx), ALLOCATABLE :: xadj(:), adjncy(:)

    n = num_nodes(graph)
    partition%part_range = extent(1, num_partitions)
    ALLOCATE(partition%assigned(1:n))
    IF (ppm_idx == i4) THEN
      IF (PRESENT(edge_weights)) THEN
        CALL graph_partition_metis_base(n, graph%edges_of_vtx, &
             graph%edges, partition%assigned, &
             num_partitions, imbalance_tolerance, vertex_weights, edge_weights)
      ELSE
        CALL graph_partition_metis_base(n, graph%edges_of_vtx, &
             graph%edges, partition%assigned, &
             num_partitions, imbalance_tolerance, vertex_weights)
      END IF
    ELSE
      ALLOCATE(xadj(SIZE(graph%edges_of_vtx)), adjncy(SIZE(graph%edges)))
      xadj = INT(graph%edges_of_vtx, ppm_idx)
      adjncy = INT(graph%edges, ppm_idx)
      IF (PRESENT(edge_weights)) THEN
        CALL graph_partition_metis_base(n, xadj, adjncy, partition%assigned, &
             num_partitions, imbalance_tolerance, vertex_weights, edge_weights)
      ELSE
        CALL graph_partition_metis_base(n, xadj, adjncy, partition%assigned, &
             num_partitions, imbalance_tolerance, vertex_weights)
      END IF
    END IF
  END SUBROUTINE graph_partition_metis_csr_mweights

END MODULE ppm_graph_partition_serial
!
! Local Variables:
! license-project-url: "https://www.dkrz.de/redmine/projects/show/scales-ppm"
! license-default: "bsd"
! End:
